#include <u.h>
#include <libc.h>
#include "iwl.h"

static int eetbloff[] = {
	[EEtblhost] 	2*0x64,
	[EEtblgen]		2*0x65,
	[EEtblreg]		2*0x66,
	[EEtblcal]		2*0x67,
	[EEtbladj]		2*0x68,
	[EEtbloth]		2*0x69
};

enum {
	EEmacaddr	= 2*0x15,

	EEchnlcnt		= 56,		/* number of channels */
};

static int eeband1[] = {				/* 2.4 GHz */
	1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14
};

static int eeband2[] = {		/* 4915-5080 MHz */
	183, 184, 185, 187, 188, 189, 192, 196, 7, 8, 11, 12, 16
};

static int eeband3[] = {		/* 5170-5320 MHz */
	34, 36, 38, 40, 42, 44, 46, 48, 52, 56, 60, 64
};

static int eeband4[] = {		/* 5500-5700 MHz */
	100, 104, 108, 112, 116, 120, 124, 128, 132, 136, 140
};

static int eeband5[] = {		/* 5725-5825 MHz */
	145, 149, 153, 157, 161, 165
};

static int eeband6[] = {		/* 2.4 GHz fat */
	1, 2, 3, 4, 5, 6, 7
};

static int eeband7[] = {		/* 5 GHz fat */
	36, 44, 52, 60, 100, 108, 116, 124, 132, 149, 157
};

struct {
	int	band;
	int	eeoffset;
	int	nchnls;
	int*	chnls;
} eechnls[] = {
	{ Band2G, 0x08 | (EEtblreg<<16), nelem(eeband1), eeband1 },
	{ Band5G, 0x26 | (EEtblreg<<16), nelem(eeband2), eeband2 },
	{ Band5G, 0x42 | (EEtblreg<<16), nelem(eeband3), eeband3 },
	{ Band5G, 0x5c | (EEtblreg<<16), nelem(eeband4), eeband4 },
	{ Band5G, 0x74 | (EEtblreg<<16), nelem(eeband5), eeband5 },
	{ Band2G, 0x82 | (EEtblreg<<16), nelem(eeband6), eeband6 },
	{ Band5G, 0x92 | (EEtblreg<<16), nelem(eeband7), eeband7 },
};

int 
eeinit(void)
{
	int efd, i, n;
	char *buf;
	Dir *d;

	buf = emalloc(strlen(iwl->path) + 10);
	sprint(buf, "%s/eeprom", iwl->path);
	d = dirstat(buf);
	if (!d) {
		free(buf);
		return 0;
	}

	iwl->eesize = d->length;
	free(iwl->eeprom);
	iwl->eeprom = emalloc(iwl->eesize);
	free(d);

	efd = open(buf, OREAD);
	free(buf);
	if (efd < 0)
		return 0;

	for(n = 0; n < iwl->eesize;) {
		i = read(efd, iwl->eeprom + n, iwl->eesize - n);
		if (i <= 0)
			break;
		n += i;
	}

	if (n != iwl->eesize) {
		werrstr("eeprom short read");
		close(efd);
		return 0;
	}

	close(efd);
	return 1;
}

int
ee16r(int addr)
{
	ushort v;

	v = *((ushort *) (iwl->eeprom + addr));
//	fprint(2, "ee16r(%ux)=%ux\n", addr, v);
	return v;
}

int
eeaddr(int offset)
{
	int val;
	int tbl;

	tbl = offset>>16;
	if (tbl > nelem(eetbloff))
		return offset;

	if (eetbloff[tbl] == 0)
		return offset;

	val = ee16r(eetbloff[tbl]) * 2;
//	fprint(2, "eeaddr: val %x\n", val);
	if (val < 0)
		return -1;

//	fprint(2, "eeadr: offset %ux base %ux ret %ux\n", offset, val, val + (offset & 0xffff));
	return val + (offset & 0xffff);
}

int
eemac(uchar *mac)
{
	ushort n;

	n = ee16r(EEmacaddr);
	mac[0] = n& 0xFF;
	mac[1] = n>>8;
	n = ee16r(EEmacaddr + 2);
	mac[2] = n&0xFF;
	mac[3] = n>>8;
	n = ee16r(EEmacaddr + 4);
	mac[4] = n&0xFF;
	mac[5] = n>>8;

	return 6;
}

int
eechnlinit(void)
{
	int i, j, n, offset;
	ushort v;
	int flags, pwr;
	Ichnl *ch;

	iwl->nchnls = EEchnlcnt;
	iwl->chnls = emalloc(iwl->nchnls * sizeof(Ichnl));
	for(n = 0, i = 0; i < 5; i++) {
		offset = eeaddr(eechnls[i].eeoffset);
//		fprint(2, "band %d offset %x\n", i, offset);
		for(j = 0; j < eechnls[i].nchnls; j++) {
			ch = &iwl->chnls[n++];
			ch->band = eechnls[i].band;
			ch->chnl = eechnls[i].chnls[j];
			v = ee16r(offset + j*2);
			pwr = v>>8;
			flags = v&0xFF;
			ch->valid = flags & 1;
			ch->ibss = flags & 2;
			ch->actvscan = flags & 8;
			ch->radar = flags & 16;
			ch->wide = flags & 32;
			ch->dfs = flags & 128;
			ch->maxpwr = pwr;

//			fprint(2, "channel %d band %s valid %d active %d\n", ch->chnl,
//				ch->band==Band2G?"2G":"5G", ch->valid, ch->actvscan);
		}
	}

	return 1;
}
