#include <u.h>
#include <libc.h>
#include <thread.h>
#include "iwl.h"

Iwl *iwl;
static Iwl globiwl;

int
iwlinit(char *path)
{
	int n;
	char *buf;

	if (path == nil)
		path = "/dev/iwl";

	n = strlen(path) + 16;
	buf = malloc(n);
	if (!buf)
		sysfatal("malloc error");

	qlock(&globiwl.lock);
	if (iwl) {
		qunlock(&globiwl.lock);
		return 1;
	}

	iwl = &globiwl;
	iwl->path = strdup(path);
	if (!eeinit())
		goto error;

	snprint(buf, n, "%s/ctl", path);
	iwl->ctlfd = open(buf, OWRITE);
	if (iwl->ctlfd  < 0) {
		fprint(2, "can't open ctl\n");
		goto error;
	}

	if (!queueinit(4)) {
		fprint(2, "can't init command queue\n");
		goto error;
	}

	if (!rxinit()) {
		fprint(2, "can't init rx\n");
		goto error;
	}

	eechnlinit();

	qunlock(&globiwl.lock);
	free(buf);
	return 1;

error:
	iwl = nil;
	qunlock(&globiwl.lock);
	free(buf);
	return 0;
}

void *
emalloc(ulong sz)
{
	void *p;

	p = malloc(sz);
	if (!p)
		sysfatal("malloc error");

	return p;
}

void *
emallocz(ulong sz, int clr)
{
	void *p;

	p = mallocz(sz, clr);
	if (!p)
		sysfatal("malloc error");

	return p;
}