enum {				/* EEPROM tables */
	EEtblhost	= 1,		/* host */
	EEtblgen	= 2,		/* general */
	EEtblreg	= 3,		/* regulatory */
	EEtblcal	= 4,		/* calibration */
	EEtbladj	= 5,		/* process adjust */
	EEtbloth	= 6,		/* others */
};

enum {
	Band2G	= 0,
	Band5G	= 1,
};

typedef struct Imsg Imsg;
typedef struct Iqueue Iqueue;
typedef struct Ichnl Ichnl;
typedef struct Iwl Iwl;

struct Imsg {
	uint		len;		/* length of the payload */
	uchar	cmd;		/* command */
	uchar	flags;	
	ushort	seq;
	uchar*	data;
	uchar*	pkt;
	Imsg*	next;

	int		datalen;	/* used to specify the "data" len of the payload */
};

struct Iqueue {
	QLock;
	int		id;
	int		txfd;
	Imsg*	txfirst;
	Imsg*	txlast;
	Rendez	txrend;
	Imsg*	rxfirst;
	Imsg*	rxlast;
	Rendez	rxrend;
	int		nexttag;
	uint		tagmap[128];
};

struct Ichnl {
	ushort	chnl;
	uchar	band;
	uchar	valid;
	uchar	actvscan;
	uchar	ibss;
	uchar	radar;
	uchar	wide;
	uchar	dfs;
	char		maxpwr;
};

struct Iwl {
	QLock	lock;
	char*	path;
	int		status;
	int		ctlfd;
	int		rxfd;

	/* eeprom */
	int		eesize;
	uchar*	eeprom;

	/* channels */
	int		nchnls;
	Ichnl*	chnls;

	/* ucode */
	int		ucinitisz;		/* init ucode instruciton size */
	uchar*	ucinitinst;		/* init ucode instructions */
	int		ucinitdsz;		/* init ucode data size */
	uchar*	ucinitdata;	/* init ucode data */
	int		ucrtisz;		/* runtime ucode instruciton size */
	uchar*	ucrtinst;		/* runtime ucode instructions */
	int		ucrtdsz;		/* runtime ucode data size */
	uchar*	ucrtdata;		/* runtime ucode data */

	Iqueue	queues[20];
};

#define p8(p, o, v)	*((uchar *) &((uchar *) (p))[o]) = (v)
#define p16(p, o, v)	*((ushort *) &((uchar *) (p))[o]) = (v)
#define p32(p, o, v)	*((uint *) &((uchar *) (p))[o]) = (v)

#define g8(p, o)	*((uchar *) &((uchar *) (p))[o])
#define g16(p, o)	*((ushort *) &((uchar *) (p))[o])
#define g32(p, o)	*((uint *) &((uchar *) (p))[o])

int iwlinit(char *);
int iwlreset(void);
void *emalloc(ulong);
void *emallocz(ulong, int);
int ctlsend(char *cmd, ...);
int eeinit(void);
int ee16r(int addr);
int eeaddr(int offset);
int eemac(uchar *macaddr);
int eechnlinit(void);
int ucinit(char *ucpath);
int ucloadinit(void);
int ucloadruntime(void);
int ucloadinit(void);

int queueinit(int qid);
int rxinit(void);
Imsg *msgalloc(int sz);
void msgfree(Imsg *);
ushort msgsend(int qid, Imsg *, int ignresp);
Imsg *msgwait(int qid, int (*)(int, Imsg *, void *), void *);

ushort cmdsend(Imsg *m, int ignresp);
Imsg *cmdwait(int (*fn)(int, Imsg *m, void *a), void *a);
Imsg *cmdrpc(Imsg *m);

Iwl *iwl;