#include <u.h>
#include <libc.h>
#include "iwl.h"

static char *Eshort = "short read";

static long
preadn(int fd, void *buf, long nbytes, vlong offset)
{
	long len, n;
	uchar *p;

	p = buf;
	len = nbytes;
	while (nbytes > 0) {
		n = pread(fd, p, len, offset);
		if (n <= 0)
			break;

		p += n;
		len -= n;
		offset += n;
	}

	return nbytes - len;
}

int
ucinit(char *ucpath)
{
	int n, m, ufd;
	uint buf[8];

	ufd = open(ucpath, OREAD);
	if (ufd < 0)
		return 0;

	n = read(ufd, buf, sizeof(buf));
	if (n != sizeof(buf)) {
		if (n >= 0)
			werrstr(Eshort);

		goto error;
	}

	iwl->ucrtisz = buf[1];
	iwl->ucrtdsz = buf[2];
	iwl->ucinitisz = buf[3];
	iwl->ucinitdsz = buf[4];

	if (iwl->ucrtisz>0x20000 || iwl->ucinitisz>0x20000) {
		werrstr("instruciton section too big");
		goto error;
	}

	if (iwl->ucrtdsz>0xC000 || iwl->ucinitdsz>0xC000) {
		werrstr("data section too big");
		goto error;
	}

	m = 24;
	iwl->ucrtinst = emalloc(iwl->ucrtisz);
	if (preadn(ufd, iwl->ucrtinst, iwl->ucrtisz, m) != iwl->ucrtisz) {
		werrstr(Eshort);
		goto error;
	}

	m += iwl->ucrtisz;
	iwl->ucrtdata = emalloc(iwl->ucrtdsz);
	if (preadn(ufd, iwl->ucrtdata, iwl->ucrtdsz, m) != iwl->ucrtdsz) {
		werrstr(Eshort);
		goto error;
	}

	m += iwl->ucrtdsz;
	iwl->ucinitinst = emalloc(iwl->ucinitisz);
	if (preadn(ufd, iwl->ucinitinst, iwl->ucinitisz, m) != iwl->ucinitisz) {
		werrstr(Eshort);
		goto error;
	}

	m += iwl->ucinitisz;
	iwl->ucinitdata = emalloc(iwl->ucinitdsz);
	if (preadn(ufd, iwl->ucinitdata, iwl->ucinitdsz, m) != iwl->ucinitdsz) {
		werrstr(Eshort);
		goto error;
	}

	close(ufd);
	return 1;

error:
	close(ufd);
	return 0;
}

static int
ucloadsection(char *sect, uchar *data, int datalen)
{
	int n, ufd;
	char *path;

	path = emalloc(strlen(iwl->path) + strlen(sect) + 8);
	sprint(path, "%s/%s", iwl->path, sect);
	ufd = open(path, OWRITE | OTRUNC);
	free(path);
	if (ufd < 0)
		return 0;

	n = write(ufd, data, datalen);
	close(ufd);
	return n == datalen;	
}

int
ucloadinit(void)
{
	if (!ucloadsection("ucinst", iwl->ucinitinst, iwl->ucinitisz))
		return 0;

	if (!ucloadsection("ucdata", iwl->ucinitdata, iwl->ucinitdsz))
		return 0;

	return 1;
}

int
ucloadruntime(void)
{
	if (!ucloadsection("ucinst", iwl->ucrtinst, iwl->ucrtisz))
		return 0;

	if (!ucloadsection("ucdata", iwl->ucrtdata, iwl->ucrtdsz))
		return 0;

	return 1;
}
