#include <u.h>
#include <libc.h>
#include <thread.h>
#include <iwl.h>

static int
staremoveall(void)
{
	Imsg *m;

	m = msgalloc(0);
	m->cmd = 0x1a;			/* remove all sta */
	m = cmdrpc(m);
	free(m);
	return 1;
}

static int
staaddbcast(void)
{
	int i;
	uchar *p;
	Imsg *m;

	m = msgalloc(0x5c);
	m->cmd = 0x18;
	p = m->data;
	p8(p, 0, 0);				/* add station */
	memset(p+4, 0xff, 6);		/* address */
	p8(p, 0x0c, 15);			/* station id */
	p8(p, 0x4c, 0);				/* station flags */
	m = cmdrpc(m);
	free(m);

	m = msgalloc(0x58);
	m->cmd = 0x4e;			/* TX link quality */
	p = m->data;
	p8(p, 0, 0x0f);				/* staid */
	p8(p, 6, 2);				/* single ant */
	p8(p, 7, 3);				/* dual ant */
	p8(p, 0xe, 3);				/* agg start thresh */
	p16(p, 0xc, 4000);

	/* 1M rate */
	for(i = 0; i < 16; i++)
		p32(p, 0x14 + i*4, 0x820A);

	m = cmdrpc(m);
	free(m);

	return 1;
}

static int
cmdrxon(void)
{
	uchar *p;
	Imsg *m;

	m = msgalloc(0x32);
	m->cmd = 0x10;
	p = m->data;
	eemac(p);
	eemac(&p[0x10]);
	p16(p, 0x1a, 0x2806);
	p8(p, 0x1c, 0xff);
	p8(p, 0x1d, 0x0f);
	p32(p, 0x20, 0x40008005);
	p32(p, 0x24, 4);
	p16(p, 0x28, 1);
	p8(p, 0x2a, 0xff);
	p8(p, 0x2b, 0xff);

	m = cmdrpc(m);
	free(m);
	return 1;
}

static int
cmdleds(void)
{
	uchar *p;
	Imsg *m;

	m = msgalloc(8);
	m->cmd = 0x48;
	p = m->data;
	p16(p, 0, 0x3e8);
	p8(p, 4, 2);
	p16(p, 6, 0x0b);

	m = cmdrpc(m);
	free(m);
	return 1;
}

static int
cmdtxpwrdbm(void)
{
	uchar *p;
	Imsg *m;

	m = msgalloc(4);
	m->cmd = 0x98;
	p = m->data;
	p8(p, 0, 0x20);
	p8(p, 1, 0x40);
	p8(p, 1, 0x7f);

	m = cmdrpc(m);
	free(m);
	return 1;
}

static int
cmdtxpwrtbl(void)
{
	uchar *p;
	Imsg *m;

	m = msgalloc(36);
	m->cmd = 0x77;
	p = m->data;
	p8(p, 0, 0x0);

	m = cmdrpc(m);
	free(m);
	return 1;
}

static Imsg *
scancreate(int band)
{
	int i, t, nchnls;
	uchar *p;
	Imsg *m;

	m = msgalloc(1024);
	p = m->data;
	memset(p, 0, 1024);
	p16(p, 4, 0x0A);
	p16(p, 6, 1);
	p32(p, 8, band==Band2G?0:1);
	p32(p, 0xa, band==Band2G?0x3BF:0x3FF);
	p32(p, 0x14, band==Band2G?5:0);
	p32(p, 0x18, 0x44);

	/* probe txcmd */
	p16(p, 0x1c, 0x46);				/* packet length */
	p32(p, 0x20, 0x2000);
	p32(p, 0x28, band==Band2G?0x420A:0x420D);
	p8(p, 0x2c, 0x0f);				/* bcast STA ID */

	/* probe frame */
	t = 0x2fc; // 0x76;
	p16(p, t+0x00, 0x40);
	memset(&p[t+0x04], 0xFF, 6);		/* bcast addr */
	eemac(&p[t+0x0A]);				/* mac address */
	memset(&p[t+0x10], 0xFF, 6);		/* bcast addr */
	p16(p, t+0x16, 0);				/* seqctl */
	p8(p, t+0x18, 0);				/* WLAN_EID_SSID */
	p8(p, t+0x19, 0);
	p8(p, t+0x1a, 1);				/* WLAN_EID_SUPP_RATES */
	p8(p, t+0x1b, 0x08);				/* number of rates */
	p8(p, t+0x1c, 0x82);				/* 1M */
	p8(p, t+0x1d, 0x84);				/* 2M */
	p8(p, t+0x1e, 0x0b);				/* 5M */
	p8(p, t+0x1f, 0x16);				/* 11M */
	p8(p, t+0x20, 0x8c);				/* 6M */
	p8(p, t+0x21, 0x98);				/* 12M */
	p8(p, t+0x22, 0xb0);				/* 24M */
	p8(p, t+0x23, 0x12);				/* 9M */
	p8(p, t+0x24, 0x32);				/* WLAN_EID_EXT_SUPP_RATES */
	p8(p, t+0x25, 0x04);				/* number of rates */
	p8(p, t+0x26, 0x24);				/* 18M */
	p8(p, t+0x27, 0x48);				/* 36M */
	p8(p, t+0x28, 0x60);				/* 48M */
	p8(p, t+0x29, 0x6c);				/* 54M */
	p8(p, t+0x2a, 0x2d);				/* WLAN_EID_HT_CAPABILITY */
	p8(p, t+0x2b, 0x1a);				/* length */
	p16(p, t+0x2c, 0x87e);			/* HT cap info */
	p8(p, t+0x2e, 0x17);				/* AMPDU params */
	/* supp mcs set */
	p8(p, t+0x2f, 0xFF);
	p8(p, t+0x30, 0xFF);
	p8(p, t+0x33, 0x01);	
	p8(p, t+0x38, 0x00);
	p8(p, t+0x39, 0x2c);
	p8(p, t+0x3a, 0x01);
	p8(p, t+0x3b, 0x03);

	/* scan channels */
	t += 0x46;
	fprint(2, "channel offset %x\n", t);
	for(nchnls = 0, i = 0; i < iwl->nchnls; i++) {
		if (!iwl->chnls[i].valid)
			continue;

//		fprint(2, "channel %d band %d pband %d\n", iwl->chnls[i].chnl,
//			iwl->chnls[i].band, band);

		if (iwl->chnls[i].band != band)
			continue;

		if (!iwl->chnls[i].actvscan)
			continue;

		if (t+12 > 1024)
			break;

		fprint(2, "channel %d\n", iwl->chnls[i].chnl);
		p32(p, t+0x00, 6 | (iwl->chnls[i].actvscan?1:0));
		p16(p, t+0x04, iwl->chnls[i].chnl);
		p8(p, t+0x06, band==Band2G?0x28:0x3b);
		p8(p, t+0x07, 0x6e);
		p16(p, t+0x08, band==Band2G?0x27:0x1a);
		p16(p, t+0x0a, band==Band2G?0x78:0x6e);
		t += 12;
		nchnls++;
	}

	m->cmd = 0x80;
	m->len = 0x46 + nchnls*12 + 0x2fc;
	p16(p, 0, m->len);		/* len */
	p8(p, 3, nchnls);

	return m;
}

void
handlebeacon(Imsg *m, uchar *p)
{
	uchar eid, elen, c;
	uchar *ep;

	ep = m->data + m->len;
	p += 24;		/* header */
	p += 8+2+2;	/* timestamp, beacon_int, capab_info */
	while (p < ep) {
		eid = *(p++);
		elen = *(p++);
		switch (eid) {
		case 0:	/* SSID */
			c = p[elen];
			p[elen] = '\0';
			fprint(2, "SSID: %s\n", (char *) p);
			p[elen] = c;
			break;
		}

		p += elen;
	}
		
}

void
handlerx(Imsg *m)
{
	uchar ftype, fsub;
	ushort len, fctl;
	uchar *p;

	p = m->data;

	/* MPDU res start */
	len = g16(p, 0);
	p += 4;

	/* frame header */
	fctl = g16(p, 0);
	ftype = (fctl>>2) & 3;
	fsub = (fctl>>4) & 0xf;
	switch (ftype) {
	case 0:		/* management frame */
		if (fsub==8)
			handlebeacon(m, p);
		else
			fprint(2, "management frame %x\n", fsub);
		break;

	case 1:		/* control frame */
		fprint(2, "control frame %x\n", fsub);
		break;

	case 2:		/* data frame */
		fprint(2, "data frame %x\n", fsub);
		break;
	}	
}

int scancheck(int q, Imsg *m, void *a)
{
	USED(q);
	USED(m);
	USED(a);
	return 1;
}

int
iwlscan(int band)
{
	int i;
	uchar cmd;
	ushort seq;
	Imsg *m;

	m = scancreate(band);
	fprint(2, "scan cmd: ");
	for(i = 0; i < m->len; i++) {
		if (i%16 == 12)
			fprint(2, "\n");

		fprint(2, "%02x ", m->data[i]);
	}
	fprint(2, "\n");

	seq = cmdsend(m, 0);
	while (1) {
		m = cmdwait(scancheck, &seq);
		if (m->cmd == 0xc1)
			handlerx(m);
		else
			fprint(2, "msg cmd %x len %d\n", m->cmd, m->len);

#ifdef notdef
		for(i = 0; i < m->len; i++) {
			if (i%16 == 12)
				fprint(2, "\n");

			fprint(2, "%02x ", m->data[i]);
		}
		fprint(2, "\n");
#endif

		cmd = m->cmd;
		free(m);
		if (cmd == 0x84)
			break;
	}

	return 1;
}

void
threadmain(int argc, char *argv[])
{
	USED(argc);
	USED(argv);
	if (!iwlinit("#w")) {
		fprint(2, "can't initialize iwl\n");
		threadexitsall("error");
	}

	if (!ucinit("iwlwifi-5000-1.ucode")) {
		fprint(2, "firmware load failed\n");
		threadexitsall("uc error");
	}

	if (!iwlreset())
		threadexitsall("fail");

	cmdrxon();
	staremoveall();
	staaddbcast();

	iwlscan(Band2G);

	threadexitsall("success");
}